import io
import os
import logging
import warnings
from math import ceil
from contextlib import closing

from google.cloud.storage import Client
from google.oauth2 import service_account
import google.auth
import gs_chunked_io as gscio
import google.cloud.storage.bucket as GSBucket

from terra_notebook_utils import WORKSPACE_BUCKET, TERRA_DEPLOYMENT_ENV, MULTIPART_THRESHOLD, IO_CONCURRENCY
from terra_notebook_utils.progress import ProgressBar


logging.getLogger("google.resumable_media.requests.download").setLevel(logging.WARNING)
logging.getLogger("gs_chunked_io.writer").setLevel(logging.WARNING)

logger = logging.getLogger(__name__)


# Suppress the annoying google gcloud _CLOUD_SDK_CREDENTIALS_WARNING warnings
warnings.filterwarnings("ignore", "Your application has authenticated using end user credentials")


def get_access_token():
    """
    Retrieve the access token using the default GCP account
    returns the same result as `gcloud auth print-access-token`
    """
    if os.environ.get("TERRA_NOTEBOOK_GOOGLE_ACCESS_TOKEN"):
        token = os.environ['TERRA_NOTEBOOK_GOOGLE_ACCESS_TOKEN']
    elif os.environ.get("GOOGLE_APPLICATION_CREDENTIALS"):
        from oauth2client.service_account import ServiceAccountCredentials
        scopes = ['https://www.googleapis.com/auth/userinfo.profile', 'https://www.googleapis.com/auth/userinfo.email']
        creds = ServiceAccountCredentials.from_json_keyfile_name(os.environ['GOOGLE_APPLICATION_CREDENTIALS'],
                                                                 scopes=scopes)
        token = creds.get_access_token().access_token
    else:
        import google.auth.transport.requests
        creds, projects = google.auth.default()
        creds.refresh(google.auth.transport.requests.Request())
        token = creds.token
    return token

def reset_bond_cache():
    import requests
    token = get_access_token()
    headers = {
        'authorization': f"Bearer {token}",
        'content-type': "application/json"
    }
    resp = requests.delete(f"http://broad-bond-{TERRA_DEPLOYMENT_ENV}.appspot.com/api/link/v1/fence/", headers=headers)
    print(resp.content)

def get_client(credentials_data: dict=None, project: str=None):
    kwargs = dict()
    if credentials_data is not None:
        creds = service_account.Credentials.from_service_account_info(credentials_data)
        kwargs['credentials'] = creds
    if project is not None:
        kwargs['project'] = project
    client = Client(**kwargs)
    if credentials_data is None:
        client._credentials.refresh(google.auth.transport.requests.Request())
    return client

def oneshot_copy(src_bucket: GSBucket, dst_bucket: GSBucket, src_key: str, dst_key: str):
    """
    Download an object into memory from `src_bucket` and upload it to `dst_bucket`
    """
    logger.info(f"Copying gs://{src_bucket.name}/{src_key} to gs://{dst_bucket.name}/{dst_key}")
    fh = io.BytesIO()
    src_bucket.blob(src_key).download_to_file(fh)
    fh.seek(0)
    dst_bucket.blob(dst_key).upload_from_file(fh)
    assert src_bucket.get_blob(src_key).crc32c == dst_bucket.get_blob(dst_key).crc32c

def multipart_copy(src_bucket: GSBucket, dst_bucket: GSBucket, src_key: str, dst_key: str):
    """
    Download/upload an object in parts from `src_bucket` to `dst_bucket`
    """
    src_blob = src_bucket.get_blob(src_key)
    print(f"Copying gs://{src_bucket.name}/{src_key} to gs://{dst_bucket.name}/{dst_key}")
    number_of_chunks = ceil(src_blob.size / gscio.reader.default_chunk_size)
    progress_bar = ProgressBar(1 + number_of_chunks,
                               prefix="Copying:",
                               size=src_blob.size // 1024 ** 2,
                               units="MB")
    with closing(progress_bar):
        with gscio.AsyncPartUploader(dst_key, dst_bucket, threads=IO_CONCURRENCY) as writer:
            for chunk_number, chunk in gscio.for_each_chunk_async(src_blob, threads=IO_CONCURRENCY):
                writer.put_part(chunk_number, chunk)
                progress_bar.update()
        progress_bar.update()
    assert src_bucket.get_blob(src_key).crc32c == dst_bucket.get_blob(dst_key).crc32c

def copy(src_bucket: GSBucket, dst_bucket: GSBucket, src_key: str, dst_key: str):
    src_blob = src_bucket.get_blob(src_key)
    if MULTIPART_THRESHOLD >= src_blob.size:
        oneshot_copy(src_bucket, dst_bucket, src_key, dst_key)
    else:
        multipart_copy(src_bucket, dst_bucket, src_key, dst_key)

def list_bucket(prefix="", bucket=WORKSPACE_BUCKET):
    for blob in get_client().bucket(bucket).list_blobs(prefix=prefix):
        yield blob.name



import os
import json
import typing
import argparse

from terra_notebook_utils import table
from terra_notebook_utils.cli import dispatch, Config


table_cli = dispatch.group("table", help=table.__doc__, arguments={
    "--workspace": dict(
        type=str,
        default=None,
        help="workspace name. If not provided, the configured CLI workspace will be used"
    ),
    "--namespace": dict(
        type=str,
        default=None,
        help=("workspace namespace (google billing project). If not provided, the configured CLI google billing "
              "project will be used.")
    ),
})


@table_cli.command("list")
def list_tables(args: argparse.Namespace):
    """
    List all tables, with column headers, in the workspace
    """
    args.workspace, args.namespace = Config.resolve(args.workspace, args.namespace)
    out = dict()
    for t, attributes in table.list_tables(args.workspace, args.namespace):
        out[t] = attributes
    print(json.dumps(out, indent=2))

@table_cli.command("get", arguments={
    "--table": dict(type=str, required=True, help="table name")
})
def get_table(args: argparse.Namespace):
    """
    Get all rows
    """
    args.workspace, args.namespace = Config.resolve(args.workspace, args.namespace)
    for e in table.list_entities(args.table, args.workspace, args.namespace):
        data = e['attributes']
        data[f'{args.table}_id'] = e['name']
        print(json.dumps(data, indent=2))

@table_cli.command("get-row", arguments={
    "--table": dict(type=str, required=True, help="table name"),
    "--id": dict(type=str, required=True, help="table name"),
})
def get_row(args: argparse.Namespace):
    """
    Get one row
    """
    args.workspace, args.namespace = Config.resolve(args.workspace, args.namespace)
    e = table.get_row(args.table, args.id, args.workspace, args.namespace)
    data = e['attributes']
    data[f'{args.table}_id'] = e['name']
    print(json.dumps(data, indent=2))

@table_cli.command("get-cell", arguments={
    "--table": dict(type=str, required=True, help="table name"),
    "--id": dict(type=str, required=True, help="id of entity"),
    "--column": dict(type=str, required=True, help="column name"),
})
def get_cell(args: argparse.Namespace):
    """
    Get cell value
    """
    args.workspace, args.namespace = Config.resolve(args.workspace, args.namespace)
    for e in table.list_entities(args.table, args.workspace, args.namespace):
        if args.id == e['name']:
            print(e['attributes'][args.column])

@table_cli.command("put-row", arguments={
    "--table": dict(type=str, required=True, help="table name"),
    "--id": dict(type=str, required=True, help="Row id. This should be unique for the table"),
    "data": dict(type=str, nargs="+", help="list of key-value pairs")
})
def put_row(args: argparse.Namespace):
    """
    Put a row.
    Example:
    tnu table put-row \\
    --table abbrv_merge_input \\
    --id 1 \\
    bucket=fc-9169fcd1-92ce-4d60-9d2d-d19fd326ff10 \\
    input_keys=test_vcfs/a.vcf.gz,test_vcfs/b.vcf.gz \\
    output_key=foo.vcf.gz
    """
    args.workspace, args.namespace = Config.resolve(args.workspace, args.namespace)
    headers = [f"{args.table}_id"]
    values = [args.id]
    for pair in args.data:
        key, val = pair.split("=")
        headers.append(key)
        values.append(val)
    tsv = "\t".join(headers) + os.linesep + "\t".join(values)
    table.upload_entities(tsv, args.workspace, args.namespace)



"""
Terra data table commands
"""
from firecloud import fiss
from typing import Optional

from terra_notebook_utils import WORKSPACE_GOOGLE_PROJECT, WORKSPACE_NAME

def _iter_table(table: str,
                workspace_name: Optional[str]=WORKSPACE_NAME,
                workspace_google_project: Optional[str]=WORKSPACE_GOOGLE_PROJECT):
    resp = fiss.fapi.get_entities(workspace_google_project, workspace_name, table)
    if 200 != resp.status_code:
        print(resp.content)
        raise Exception(f"Expected status 200, got {resp.status_code}")
    for item in resp.json():
        yield item

def _get_item_val(item: dict, key: str):
    if "name" == key:
        return item['name']
    else:
        return item['attributes'][key]

def fetch_attribute(table: str, filter_column: str, filter_val: str, attribute: str):
    """
    Fetch `attribute` from `table` from the same row containing `filter_val` in column `filter_column`
    """
    for item in _iter_table(table):
        if filter_val == _get_item_val(item, filter_column):
            return _get_item_val(item, attribute)
    else:
        raise ValueError(f"No row found for table {table}, filter_column {filter_column} filter_val {filter_val}")

def fetch_object_id(table: str, file_name: str):
    """
    Fetch `object_id` associated with `file_name` from `table`.
    DRS urls, when available, are stored in `object_id`.
    """
    return fetch_attribute(table, "file_name", file_name, "object_id")

def fetch_drs_url(table: str, file_name: str):
    val = fetch_object_id(table, file_name)
    if not val.startswith("drs://"):
        raise ValueError(f"Expected DRS url in {table} for {file_name}, got {val} instead.")
    return val

def list_tables(workspace_name: Optional[str]=WORKSPACE_NAME,
                workspace_google_project: Optional[str]=WORKSPACE_GOOGLE_PROJECT):
    resp = fiss.fapi.list_entity_types(workspace_google_project, workspace_name)
    resp.raise_for_status()
    for ent_type, data in resp.json().items():
        yield ent_type, data['attributeNames']

def list_entities(ent_type: str,
                  workspace_name: Optional[str]=WORKSPACE_NAME,
                  workspace_google_project: Optional[str]=WORKSPACE_GOOGLE_PROJECT):
    resp = fiss.fapi.get_entities(workspace_google_project, workspace_name, ent_type)
    resp.raise_for_status()
    for ent in resp.json():
        yield ent

def get_row(table: str,
            entity_id: str,
            workspace_name: Optional[str]=WORKSPACE_NAME,
            workspace_google_project: Optional[str]=WORKSPACE_GOOGLE_PROJECT):
    resp = fiss.fapi.get_entity(workspace_google_project, workspace_name, table, entity_id)
    resp.raise_for_status()
    return resp.json()

def delete_entities(entities: list):
    ents = [dict(entityType=e['entityType'], entityName=e['name'])
            for e in entities]
    resp = fiss.fapi.delete_entities(WORKSPACE_GOOGLE_PROJECT, WORKSPACE_NAME, ents)
    resp.raise_for_status()

def delete_table(ent_type: str):
    delete_entities([e for e in list_entities(ent_type)])

def upload_entities(tsv_data,
                    workspace_name: Optional[str]=WORKSPACE_NAME,
                    workspace_google_project: Optional[str]=WORKSPACE_GOOGLE_PROJECT):
    resp = fiss.fapi.upload_entities(workspace_google_project, workspace_name, tsv_data, model="flexible")
    resp.raise_for_status()

def print_column(table: str, column: str):
    for item in _iter_table(table):
        print(_get_item_val(item, column))



import os

WORKSPACE_NAME = os.environ.get('WORKSPACE_NAME', None)
WORKSPACE_GOOGLE_PROJECT = os.environ.get('GOOGLE_PROJECT')  # This env var is set in Terra notebooks
TERRA_DEPLOYMENT_ENV = os.environ.get('TERRA_DEPLOYMENT_ENV', 'prod')
MARTHA_URL_VERSION = os.environ.get('MARTHA_URL_VERSION', 'martha_v3')

if not WORKSPACE_GOOGLE_PROJECT:
    WORKSPACE_GOOGLE_PROJECT = os.environ.get('GCP_PROJECT')  # Useful for running outside of notebook
if not WORKSPACE_GOOGLE_PROJECT:
    WORKSPACE_GOOGLE_PROJECT = os.environ.get('GCLOUD_PROJECT')  # Fallback
# For a list of gcloud project related environment variables, see:
# https://cloud.google.com/functions/docs/env-var#environment_variables_set_automatically
WORKSPACE_BUCKET = os.environ.get('WORKSPACE_BUCKET', None)
_GS_SCHEMA = 'gs://'
if WORKSPACE_BUCKET is not None and WORKSPACE_BUCKET.startswith(_GS_SCHEMA):
    WORKSPACE_BUCKET = WORKSPACE_BUCKET[len(_GS_SCHEMA):]  # Chop off the bucket schema

MULTIPART_THRESHOLD = 1024 * 1024 * 32
IO_CONCURRENCY = 3

MARTHA_URL = f"https://us-central1-broad-dsde-{TERRA_DEPLOYMENT_ENV}.cloudfunctions.net/{MARTHA_URL_VERSION}"

from terra_notebook_utils import drs, profile, table, vcf, workspace



import json
import typing
import argparse

from terra_notebook_utils import vcf
from terra_notebook_utils.cli import dispatch, Config
import google.cloud.storage.blob


vcf_cli = dispatch.group("vcf", help=vcf.__doc__, arguments={
    "path": dict(
        help="local path, gs://, or drs://"
    ),
    "--google-billing-project": dict(
        type=str,
        required=False,
        default=Config.info['workspace_google_project'],
        help=("The billing project for GS requests. "
              "If omitted, the CLI configured `workspace_google_project` will be used. "
              "Note that DRS URLs also involve a GS request.")
    ),
})


@vcf_cli.command("head")
def head(args: argparse.Namespace):
    """
    Output VCF header.
    """
    _, args.google_billing_project = Config.resolve(None, args.google_billing_project)
    blob = _get_blob(args.path, args.google_billing_project)
    if blob:
        info = vcf.VCFInfo.with_blob(blob)
    else:
        info = vcf.VCFInfo.with_file(args.path)
    info.print_header()

@vcf_cli.command("samples")
def samples(args: argparse.Namespace):
    """
    Output VCF samples.
    """
    _, args.google_billing_project = Config.resolve(None, args.google_billing_project)
    blob = _get_blob(args.path, args.google_billing_project)
    if blob:
        info = vcf.VCFInfo.with_blob(blob)
    else:
        info = vcf.VCFInfo.with_file(args.path)
    print(json.dumps(info.samples, indent=2))

@vcf_cli.command("stats")
def stats(args: argparse.Namespace):
    """
    Output VCF stats.
    """
    _, args.google_billing_project = Config.resolve(None, args.google_billing_project)
    blob = _get_blob(args.path, args.google_billing_project)
    if blob:
        info = vcf.VCFInfo.with_blob(blob)
        size = blob.size
    else:
        import os
        info = vcf.VCFInfo.with_file(args.path)
        size = os.path.getsize(os.path.abspath(args.path))
    stats = {
        'first data line chromosome': info.chrom,
        'length associated with first data line chromosome': info.length,
        'number of samples': len(info.samples),
        'size': size
    }
    print(json.dumps(stats, indent=2))

def _get_blob(path: str, google_project: str) -> google.cloud.storage.blob:
    if path.startswith("gs://"):
        from terra_notebook_utils import gs
        path = path.split("gs://", 1)[1]
        bucket_name, key = path.split("/", 1)
        blob = gs.get_client(project=google_project).bucket(bucket_name).get_blob(key)
    elif path.startswith("drs://"):
        from terra_notebook_utils import drs
        client, info = drs.resolve_drs_for_gs_storage(path)
        blob = client.bucket(info.bucket_name, user_project=google_project).get_blob(info.key)
    else:
        blob = None
    return blob



"""
Utilities for working with DRS objects
"""
import os
import sys
import re
import json
import logging
import requests
import traceback
from concurrent.futures import ThreadPoolExecutor
from functools import lru_cache
from collections import namedtuple
from typing import Tuple, Iterable, Optional, Callable, IO
from google.cloud.exceptions import NotFound, Forbidden

from terra_notebook_utils import (WORKSPACE_GOOGLE_PROJECT, WORKSPACE_BUCKET, WORKSPACE_NAME, MULTIPART_THRESHOLD,
                                  IO_CONCURRENCY, MARTHA_URL)
from terra_notebook_utils import gs, tar_gz, TERRA_DEPLOYMENT_ENV, _GS_SCHEMA

import gs_chunked_io as gscio
from gs_chunked_io import async_collections


logger = logging.getLogger(__name__)

DRSInfo = namedtuple("DRSInfo", "credentials bucket_name key name size updated")

class GSBlobInaccessible(Exception):
    pass

class DRSResolutionError(Exception):
    pass

def _parse_gs_url(gs_url: str) -> Tuple[str, str]:
    if gs_url.startswith(_GS_SCHEMA):
        bucket_name, object_key = gs_url[len(_GS_SCHEMA):].split("/", 1)
        return bucket_name, object_key
    else:
        raise RuntimeError(f'Invalid gs url schema.  {gs_url} does not start with {_GS_SCHEMA}')

@lru_cache()
def enable_requester_pays(workspace_name: Optional[str]=WORKSPACE_NAME,
                          google_billing_project: Optional[str]=WORKSPACE_GOOGLE_PROJECT):
    assert workspace_name
    import urllib.parse
    encoded_workspace = urllib.parse.quote(workspace_name)
    rawls_url = (f"https://rawls.dsde-{TERRA_DEPLOYMENT_ENV}.broadinstitute.org/api/workspaces/"
                 f"{google_billing_project}/{encoded_workspace}/enableRequesterPaysForLinkedServiceAccounts")
    logger.info("Enabling requester pays for your workspace. This will only take a few seconds...")
    access_token = gs.get_access_token()

    headers = {
        'authorization': f"Bearer {access_token}",
        'content-type': "application/json"
    }
    resp = requests.put(rawls_url, headers=headers)

    if resp.status_code != 204:
        logger.warning(f"Failed to init requester pays for workspace {WORKSPACE_GOOGLE_PROJECT}/{WORKSPACE_NAME}.")
        logger.warning("You will not be able to access drs urls that interact with requester pays buckets.")

def fetch_drs_info(drs_url: str) -> dict:
    """
    Request DRS infromation from martha.
    """
    access_token = gs.get_access_token()

    headers = {
        'authorization': f"Bearer {access_token}",
        'content-type': "application/json"
    }

    logger.info(f"Resolving DRS uri '{drs_url}' through '{MARTHA_URL}'.")

    resp = requests.post(MARTHA_URL, headers=headers, data=json.dumps(dict(url=drs_url)))

    if 200 == resp.status_code:
        resp_data = resp.json()
    else:
        logger.warning(resp.content)
        response_json = resp.json()

        if 'response' in response_json:
            if 'text' in response_json['response']:
                error_details = f"Error: {response_json['response']['text']}"
            else:
                error_details = ""
        else:
            error_details = ""

        raise DRSResolutionError(f"Unexpected response while resolving DRS path. Expected status 200, got "
                                 f"{resp.status_code}. {error_details}")

    return resp_data

def extract_credentials_from_drs_response(response: dict) -> Optional[dict]:
    if 'googleServiceAccount' not in response or response['googleServiceAccount'] is None:
        credentials_data = None
    else:
        credentials_data = response['googleServiceAccount']['data']

    return credentials_data

def convert_martha_v2_response_to_DRSInfo(drs_url: str, drs_response: dict) -> DRSInfo:
    """
    Convert response from martha_v2 to DRSInfo
    """
    if 'data_object' in drs_response['dos']:
        credentials_data = extract_credentials_from_drs_response(drs_response)
        data_object = drs_response['dos']['data_object']

        if 'urls' not in data_object:
            raise DRSResolutionError(f"No GS url found for DRS uri '{drs_url}'")
        else:
            data_url = None
            for url_info in data_object['urls']:
                if 'url' in url_info and url_info['url'].startswith(_GS_SCHEMA):
                    data_url = url_info['url']
                    break
            if data_url is None:
                raise DRSResolutionError(f"No GS url found for DRS uri '{drs_url}'")

        bucket_name, key = _parse_gs_url(data_url)
        return DRSInfo(credentials=credentials_data,
                       bucket_name=bucket_name,
                       key=key,
                       name=data_object.get('name'),
                       size=data_object.get('size'),
                       updated=data_object.get('updated'))
    else:
        raise DRSResolutionError(f"No metadata was returned for DRS uri '{drs_url}'")

def convert_martha_v3_response_to_DRSInfo(drs_url: str, drs_response: dict) -> DRSInfo:
    """
    Convert response from martha_v3 to DRSInfo
    """
    if 'gsUri' not in drs_response:
        raise DRSResolutionError(f"No GS url found for DRS uri '{drs_url}'")

    credentials_data = extract_credentials_from_drs_response(drs_response)

    return DRSInfo(credentials=credentials_data,
                   bucket_name=drs_response.get('bucket'),
                   key=drs_response.get('name'),
                   name=drs_response.get('fileName'),
                   size=drs_response.get('size'),
                   updated=drs_response.get('timeUpdated'))

def resolve_drs_info_for_gs_storage(drs_url: str) -> DRSInfo:
    """
    Attempt to resolve gs:// url and credentials for a DRS object.
    """
    assert drs_url.startswith("drs://")
    drs_response: dict = fetch_drs_info(drs_url)

    if 'dos' in drs_response:
        return convert_martha_v2_response_to_DRSInfo(drs_url, drs_response)
    else:
        return convert_martha_v3_response_to_DRSInfo(drs_url, drs_response)

def resolve_drs_for_gs_storage(drs_url: str) -> Tuple[gs.Client, DRSInfo]:
    """
    Attempt to resolve gs:// url and credentials for a DRS object. Instantiate and return the Google Storage client.
    """
    assert drs_url.startswith("drs://")

    try:
        info = resolve_drs_info_for_gs_storage(drs_url)
    except DRSResolutionError:
        raise
    except Exception:
        raise

    if info.credentials is not None:
        project_id = info.credentials['project_id']
    else:
        project_id = None

    client = gs.get_client(info.credentials, project=project_id)
    return client, info

def copy_to_local(drs_url: str,
                  filepath: str,
                  workspace_name: Optional[str]=WORKSPACE_NAME,
                  google_billing_project: Optional[str]=WORKSPACE_GOOGLE_PROJECT):
    """
    Copy a DRS object to the local filesystem.
    """
    assert drs_url.startswith("drs://")
    enable_requester_pays(workspace_name, google_billing_project)
    client, info = resolve_drs_for_gs_storage(drs_url)
    blob = client.bucket(info.bucket_name, user_project=google_billing_project).blob(info.key)
    if os.path.isdir(filepath):
        filename = info.name or _url_basename(drs_url)
        filepath = os.path.join(os.path.abspath(filepath), filename)
    with open(filepath, "wb") as fh:
        logger.info(f"Downloading {drs_url} to {filepath}")
        blob.download_to_file(fh)

def head(drs_url: str,
         num_bytes: int = 1,
         buffer: int = MULTIPART_THRESHOLD,
         workspace_name: Optional[str] = WORKSPACE_NAME,
         google_billing_project: Optional[str] = WORKSPACE_GOOGLE_PROJECT):
    """
    Head a DRS object by byte.

    :param drs_url: A drs:// schema URL.
    :param num_bytes: Number of bytes to print from the DRS object.
    :param workspace_name: The name of the terra workspace.
    :param google_billing_project: The name of the terra google billing project.
    """
    assert drs_url.startswith("drs://"), f'Not a DRS schema: {drs_url}'
    enable_requester_pays(workspace_name, google_billing_project)
    try:
        client, info = resolve_drs_for_gs_storage(drs_url)
        blob = client.bucket(info.bucket_name, user_project=google_billing_project).blob(info.key)
        try:
            # sys.stdout.buffer is used outside of a python notebook since that's the standard non-lossy way
            # to write/display bytes; sys.stdout.buffer is not available inside of a python notebook
            # though, as sys.stdout is a ipykernel.iostream.OutStream object:
            # https://github.com/ipython/ipykernel/blob/master/ipykernel/iostream.py#L265
            # so we use bare sys.stdout and rely on the ipykernel method's lossy unicode stream
            stdout_buffer = getattr(sys.stdout, 'buffer', sys.stdout)
            with gscio.Reader(blob, chunk_size=buffer) as handle:
                stdout_buffer.write(handle.read(num_bytes))

        except (NotFound, Forbidden):
            raise GSBlobInaccessible(f'The DRS URL: {drs_url}\n'
                                     f'Could not be accessed because of:\n'
                                     f'{traceback.format_exc()}')
    except DRSResolutionError:
        raise GSBlobInaccessible(f'The DRS URL: {drs_url}\n Could not be accessed because of:\n'
                                 f'{traceback.format_exc()}')
    except Exception:
        raise

def copy_to_bucket(drs_url: str,
                   dst_key: str,
                   dst_bucket_name: str=None,
                   workspace_name: Optional[str]=WORKSPACE_NAME,
                   google_billing_project: Optional[str]=WORKSPACE_GOOGLE_PROJECT):
    """
    Resolve `drs_url` and copy into user-specified bucket `dst_bucket`.
    If `dst_bucket` is None, copy into workspace bucket.
    """
    assert drs_url.startswith("drs://")
    enable_requester_pays(workspace_name, google_billing_project)
    if dst_bucket_name is None:
        dst_bucket_name = WORKSPACE_BUCKET
    src_client, src_info = resolve_drs_for_gs_storage(drs_url)
    dst_client = gs.get_client()
    src_bucket = src_client.bucket(src_info.bucket_name, user_project=google_billing_project)
    dst_bucket = dst_client.bucket(dst_bucket_name)
    logger.info(f"Beginning to copy from {src_bucket} to {dst_bucket}. This can take a while for large files...")
    gs.copy(src_bucket, dst_bucket, src_info.key, dst_key)

def copy(drs_url: str,
         dst: str,
         workspace_name: Optional[str]=WORKSPACE_NAME,
         google_billing_project: Optional[str]=WORKSPACE_GOOGLE_PROJECT):
    """
    Copy a DRS object to either the local filesystem, or to a Google Storage location if `dst` starts with "gs://".
    """
    assert drs_url.startswith("drs://")
    if dst.startswith("gs://"):
        bucket_name, key = _bucket_name_and_key(dst)
        copy_to_bucket(drs_url,
                       key,
                       bucket_name,
                       workspace_name=workspace_name,
                       google_billing_project=google_billing_project)
    else:
        copy_to_local(drs_url, dst, workspace_name, google_billing_project)

def copy_batch(drs_urls: Iterable[str],
               dst: str,
               workspace_name: Optional[str]=WORKSPACE_NAME,
               google_billing_project: Optional[str]=WORKSPACE_GOOGLE_PROJECT):
    enable_requester_pays(workspace_name, google_billing_project)
    with ThreadPoolExecutor(max_workers=IO_CONCURRENCY) as oneshot_executor:
        oneshot_pool = async_collections.AsyncSet(oneshot_executor, concurrency=IO_CONCURRENCY)
        for drs_url in drs_urls:
            assert drs_url.startswith("drs://")
            src_client, src_info = resolve_drs_for_gs_storage(drs_url)
            src_bucket = src_client.bucket(src_info.bucket_name, user_project=google_billing_project)
            src_blob = src_bucket.get_blob(src_info.key)
            basename = src_info.name or _url_basename(drs_url)
            if dst.startswith("gs://"):
                if dst.endswith("/"):
                    raise ValueError("Bucket destination cannot end with '/'")
                dst_bucket_name, dst_pfx = _bucket_name_and_key(dst)
                dst_bucket = gs.get_client().bucket(dst_bucket_name)
                dst_key = f"{dst_pfx}/{basename}"
                if MULTIPART_THRESHOLD >= src_blob.size:
                    oneshot_pool.put(gs.oneshot_copy, src_bucket, dst_bucket, src_info.key, dst_key)
                else:
                    gs.multipart_copy(src_bucket, dst_bucket, src_info.key, dst_key)
            else:
                oneshot_pool.put(copy_to_local, drs_url, dst, workspace_name, google_billing_project)
        for _ in oneshot_pool.consume():
            pass

def extract_tar_gz(drs_url: str,
                   dst_pfx: str=None,
                   dst_bucket_name: str=None,
                   workspace_name: Optional[str]=WORKSPACE_NAME,
                   google_billing_project: Optional[str]=WORKSPACE_GOOGLE_PROJECT):
    """
    Extract a `.tar.gz` archive resolved by a DRS url into a Google Storage bucket.
    """
    if dst_bucket_name is None:
        dst_bucket_name = WORKSPACE_BUCKET
    assert dst_bucket_name
    enable_requester_pays(workspace_name, google_billing_project)
    src_client, src_info = resolve_drs_for_gs_storage(drs_url)
    src_bucket = src_client.bucket(src_info.bucket_name, user_project=google_billing_project)
    dst_bucket = gs.get_client().bucket(dst_bucket_name)
    with gscio.Reader(src_bucket.get_blob(src_info.key), threads=2) as fh:
        tar_gz.extract(fh, dst_bucket, root=dst_pfx)

def _url_basename(url: str) -> str:
    schema_re = "[a-z]+://"
    path_re = ".*"
    m = re.match(f"(?P<schema>{schema_re})(?P<path>{path_re})", url)
    if m:
        parts = m['path'].rsplit("/", 1)
        if 2 == len(parts):
            return parts[-1]
        else:
            raise ValueError(f"'{url}' missing basename")
    else:
        raise ValueError(f"'{url}' does not match '{schema_re}{path_re}'")

def _bucket_name_and_key(gs_url: str) -> Tuple[str, str]:
    assert gs_url.startswith("gs://")
    parts = gs_url[5:].split("/", 1)
    if 1 >= len(parts) or not parts[1]:
        raise ValueError(f"Invalid bucket location: {gs_url}"
                         "gs:// url should contain bucket name and key with '/' delimiter.")
    bucket_name, key = parts
    return bucket_name, key



import time
from contextlib import AbstractContextManager

class profile(AbstractContextManager):
    """
    Profile methods or code blocks with decorators or contexts, respectively.

    @profile("profiling my method")
    def my_method(*args, **kwargs):
       ...

    with profile("profiling my block"):
       ...
    """
    def __init__(self, name="default"):
        self.name = name

    def __enter__(self, *args, **kwargs):
        self.start = time.time()

    def __exit__(self, *args, **kwargs):
        print(self.name, time.time() - self.start)

    def __call__(self, meth):
        def wrapper(*args, **kwargs):
            start = time.time()
            res = meth(*args, **kwargs)
            dur = time.time() - start
            print(f"{self.name} took {dur} seconds")
            return res
        return wrapper



"""
Configure the CLI
"""
import typing
import argparse

from terra_notebook_utils.cli import Config, dispatch


config_cli = dispatch.group("config", help=__doc__)


@config_cli.command("set-workspace", arguments={
    "workspace": dict()
})
def set_config_workspace(args: argparse.Namespace):
    """
    Set workspace for cli commands
    """
    Config.info["workspace"] = args.workspace
    Config.write()

@config_cli.command("set-workspace-google-project", arguments={
    "billing_project": dict()
})
def set_config_billing_project(args: argparse.Namespace):
    """
    Set billing project for cli commands
    """
    Config.info["workspace_google_project"] = args.billing_project
    Config.write()

@config_cli.command("print")
def config_print(args: argparse.Namespace):
    """
    Print the CLI configuration.
    """
    import json
    print(json.dumps(Config.info, indent=2))



"""
Workspace information and operations
"""
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import List, Optional

from firecloud import fiss

from terra_notebook_utils import gs, WORKSPACE_BUCKET, WORKSPACE_NAME


def get_workspace(workspace: Optional[str]=WORKSPACE_NAME, namespace: Optional[str]=None) -> dict:
    namespace = namespace or get_workspace_namespace(workspace)
    resp = fiss.fapi.get_workspace(namespace, workspace)
    resp.raise_for_status()
    return resp.json()

def list_workspaces() -> List[dict]:
    """
    List workspaces available to current user.
    """
    resp = fiss.fapi.list_workspaces()
    resp.raise_for_status()
    return resp.json()

def get_workspace_bucket(workspace: Optional[str]=WORKSPACE_NAME) -> Optional[str]:
    """
    Get Google Storage bucket associated with a workspace.
    """
    if not workspace:
        raise ValueError("Expected workspace")
    else:
        for ws in list_workspaces():
            if ws['workspace']['name'] == workspace:
                return ws['workspace']['bucketName']
        return None

def get_workspace_namespace(workspace: Optional[str]=WORKSPACE_NAME) -> Optional[str]:
    """
    Best effort discovery of workspace namespace.
    If two namespaces share a workspace of the same name, the first namespace encountered will be returned.
    """
    if not workspace:
        raise ValueError("Expected workspace")
    else:
        for ws in list_workspaces():
            if ws['workspace']['name'] == workspace:
                return ws['workspace']['namespace']
        return None

def remove_workflow_logs(bucket_name=WORKSPACE_BUCKET, submission_id: str=None) -> List[str]:
    """
    Experimental: do not use
    """
    bucket = gs.get_client().bucket(bucket_name)

    def _is_workflow_log(blob):
        fname = blob.name.rsplit("/", 1)[-1]
        return fname.startswith("workflow") and fname.endswith(".log")

    def _delete_blob(blob):
        blob.delete()
        return blob.name

    prefixes = {blob.name.split("/", 1)[0] for blob in bucket.list_blobs()
                if _is_workflow_log(blob)}
    if submission_id is not None:
        prefixes = {pfx for pfx in prefixes
                    if pfx == submission_id}
    for pfx in prefixes:
        blobs_to_delete = [blob for blob in bucket.list_blobs(prefix=pfx)]
        print(f"Deleting {len(blobs_to_delete)} objects for {pfx}")
        with ThreadPoolExecutor(max_workers=8) as e:
            futures = [e.submit(_delete_blob, blob) for blob in blobs_to_delete]
            deleted_manifest = [f.result() for f in as_completed(futures)]

    return deleted_manifest



import sys
import time
import threading
import sched
from math import floor, ceil
from contextlib import AbstractContextManager
from concurrent.futures import ThreadPoolExecutor, as_completed


class RateLimited:
    """
    Decorator to rate limit method calls. Raises `exception` if method is called more quickly than
    `rate` times per second, or does nothing if exception is None.
    """
    def __init__(self, rate: float, exception: Exception=None):
        self.period = 1 / rate
        self.exception = exception
        self.reset()

    def reset(self):
        self._last_call = time.time() - 10 * self.period

    def __call__(self, func):
        def wrapper(*args, **kwargs):
            now = time.time()
            if now - self._last_call >= self.period:
                self._last_call = now
                return func(*args, **kwargs)
            elif self.exception is not None:
                raise self.exception("Too soon")

        wrapper.reset = self.reset
        return wrapper


class ProgressBar(AbstractContextManager):
    def __init__(self, number_of_steps, prefix="", units="", size: int=None):
        self.number_of_steps = number_of_steps
        self.steps_completed = 0
        self.prefix = prefix
        self.units = units
        self.size = size
        self._length = 40  # This seems like a reasonable length for notebook output
        self._lock = threading.Lock()
        self._start_time = time.time()

    def update(self, number_of_steps=1):
        self.steps_completed += number_of_steps
        slots_filled = int(self._length * self.steps_completed / self.number_of_steps)
        duration = ceil(time.time() - self._start_time)
        portion_complete = self.steps_completed / self.number_of_steps
        bar = "#" * slots_filled
        bar += " " * (self._length - slots_filled)
        bar += f" {floor(100 * portion_complete)}%"
        if self.size is not None:
            bar += f" of {self.size}{self.units}"
            bar += f" %.2f{self.units}/s" % (self.size * portion_complete / duration)
        self._bar = bar
        self._duration = duration
        self._print()

    @RateLimited(2.0)
    def _print(self):
        with self._lock:
            print("\r", f"{self.prefix} {self._bar} ({self._duration} seconds)", end="")

    def close(self, message=None):
        self._print.reset()
        self._print()
        with self._lock:
            print()
            if message is not None:
                print(message)

    def __exit__(self, *args, **kwargs):
        self.close()


class ProgressReporter(AbstractContextManager):
    def __init__(self, units: str="lines"):
        self.start_time = time.time()
        self._checkpoint_time = [self.start_time, self.start_time]
        self._units_processed = [0, 0]
        self._lock = threading.Lock()
        self.units = units

    @property
    def duration(self):
        return self._checkpoint_time[-1] - self.start_time

    @property
    def average_rate(self):
        return self._units_processed[-1] / self.duration

    @property
    def instantaneous_rate(self):
        _units_processed = [self._units_processed[-1] - self._units_processed[-2]]
        duration = self._checkpoint_time[-1] - self._checkpoint_time[-2]
        return _units_processed / duration

    @property
    def units_processed(self):
        return self._units_processed[-1]

    def checkpoint(self, _units_processed: int):
        self._units_processed = [self._units_processed[-1], self._units_processed[-1] + _units_processed]
        self._checkpoint_time = [self._checkpoint_time[-1], time.time()]
        self._print()

    @RateLimited(2.0)
    def _print(self):
        with self._lock:
            print("\r",
                  f"%9i {self.units} processed," % self._units_processed[-1],
                  f"%6.0f {self.units}/s" % self.average_rate,
                  "%7.2f s" % self.duration,
                  end="")

    def close(self):
        self._print.reset()
        self._print()
        with self._lock:
            print()

    def __exit__(self, *args, **kwargs):
        self.close()



"""
Configure the CLI
"""
import os
import json
from typing import Optional

import cli_builder

from terra_notebook_utils import WORKSPACE_NAME, WORKSPACE_GOOGLE_PROJECT


class Config:
    info = dict(workspace=None, workspace_google_project=None)
    path = os.path.join(os.path.expanduser("~"), ".tnu_config")

    @classmethod
    def load(cls):
        if not os.path.isfile(cls.path):
            cls.write()
        with open(cls.path) as fh:
            cls.info = json.loads(fh.read())

    @classmethod
    def write(cls):
        with open(cls.path, "w") as fh:
            fh.write(json.dumps(cls.info, indent=2))

    @classmethod
    def resolve(cls, override_workspace: Optional[str], override_namespace: Optional[str]):
        workspace = override_workspace or (cls.info['workspace'] or WORKSPACE_NAME)
        namespace = override_namespace or (cls.info['workspace_google_project'] or WORKSPACE_GOOGLE_PROJECT)
        if workspace and namespace is None:
            from terra_notebook_utils.workspace import get_workspace_namespace
            namespace = get_workspace_namespace(workspace)
        if not workspace:
            raise RuntimeError("This command requies a workspace. Either pass in a workspace with `--workspace`,"
                               " or configure a default workspace for the cli (see `tnu config --help`)."
                               " A default workspace may also be configued by setting the `WORKSPACE_NAME` env var")
        return workspace, namespace
Config.load()


dispatch = cli_builder.Dispatch(
    """
    Welcome to the terra-notebook-utils cli
         ___
       _(((,|    What's DNA??
      /  _-\\\\
     / C o\o \\
   _/_    __\ \\     __ __     __ __     __ __     __
  /   \ \___/  )   /--X--\   /--X--\   /--X--\   /--
  |    |\_|\  /   /--/ \--\ /--/ \--\ /--/ \--\ /--/
  |    |#  #|/          \__X__/   \__X__/   \__X__/
  (   /     |
   |  |#  # |
   |  |    #|
   |  | #___n_,_
,-/   7-' .     `\\
`-\...\-_   -  o /
   |#  # `---U--'
   `-v-^-'\/
     \  |_|_ Wny
     (___mnnm
    """
)



import os
import unittest


def testmode(mode_for_test):
    def wrapper(func):
        mode = os.environ.get("TNU_TESTMODE", "workspace_access")
        return unittest.skipUnless(mode_for_test in mode, f"Skipping {mode_for_test} test")(func)
    return wrapper



import glob
import os
from setuptools import setup, find_packages

install_requires = [line.rstrip() for line in open(os.path.join(os.path.dirname(__file__), "requirements.txt"))]

with open("README.md") as fh:
    long_description = fh.read()

def get_version():
    pkg_root = os.path.abspath(os.path.join(os.path.dirname(__file__)))  # noqa
    filepath = os.path.join(pkg_root, "terra_notebook_utils", "version.py")
    version = dict()
    with open(filepath) as fh:
        exec(fh.read().strip(), version)
    return version['__version__']

setup(
    name="terra-notebook-utils",
    version=get_version(),
    description="Utilities for the Terra notebook environment.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/DataBiosphere/terra-notebook-utils",
    author="Brian Hannafious",
    author_email="bhannafi@ucsc.edu",
    license="MIT",
    packages=find_packages(exclude=["tests"]),
    scripts=glob.glob("scripts/*"),
    zip_safe=False,
    install_requires=install_requires,
    platforms=["MacOS X", "Posix"],
    test_suite="test",
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7"
    ]
)
